# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class RobotAudio(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRobotAudio(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RobotAudio()
        x.Init(buf, n + offset)
        return x

    # RobotAudio
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RobotAudio
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # RobotAudio
    def AudioEventId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # RobotAudio
    def AudioEventIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # RobotAudio
    def AudioEventIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # RobotAudio
    def Volume(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 1.0

    # RobotAudio
    def Probability(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # RobotAudio
    def ProbabilityAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # RobotAudio
    def ProbabilityLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # RobotAudio
    def HasAlts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

def RobotAudioStart(builder): builder.StartObject(5)
def RobotAudioAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def RobotAudioAddAudioEventId(builder, audioEventId): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(audioEventId), 0)
def RobotAudioStartAudioEventIdVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def RobotAudioAddVolume(builder, volume): builder.PrependFloat32Slot(2, volume, 1.0)
def RobotAudioAddProbability(builder, probability): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(probability), 0)
def RobotAudioStartProbabilityVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RobotAudioAddHasAlts(builder, hasAlts): builder.PrependBoolSlot(4, hasAlts, 1)
def RobotAudioEnd(builder): return builder.EndObject()



"""

Code for reading and writing Cozmo animations in FlatBuffers .bin and JSON format.

Cozmo animations are stored in files/cozmo/cozmo_resources/assets/animations inside the Cozmo mobile application.

Animation data structures are declared in FlatBuffers format in files/cozmo/cozmo_resources/config/cozmo_anim.fbs .

"""

from typing import List, Union, Dict, TextIO, BinaryIO
from abc import ABC
import os
import json
import glob

import flatbuffers

from . import CozmoAnim


class AnimBase(ABC):
    """ Animation element base class. """

    def __init__(self):
        pass

    def to_dict(self) -> dict:
        pass

    @classmethod
    def from_dict(cls, data):
        pass

    def to_fb(self, builder: flatbuffers.Builder):
        pass

    @classmethod
    def from_fb(cls, buf):
        pass


class AnimKeyframe(AnimBase, ABC):
    """ Animation keyframe base class. """

    def __init__(self):
        super().__init__()


class AnimClip(AnimBase):
    """ Animation clip class. """

    def __init__(self, name: str, keyframes: List[AnimKeyframe] = ()):
        super().__init__()
        self.name = str(name)
        self.keyframes = list(keyframes)

    def to_dict(self) -> dict:
        data = {
            "Name": self.name,
            "keyframes": {
                "LiftHeightKeyFrame": [],
                "ProceduralFaceKeyFrame": [],
                "HeadAngleKeyFrame": [],
                "RobotAudioKeyFrame": [],
                "BackpackLightsKeyFrame": [],
                "FaceAnimationKeyFrame": [],
                "EventKeyFrame": [],
                "BodyMotionKeyFrame": [],
                "RecordHeadingKeyFrame": [],
                "TurnToRecordedHeadingKeyFrame": [],
            }
        }
        for keyframe in self.keyframes:
            keyframe_data = keyframe.to_dict()
            if isinstance(keyframe, AnimLiftHeight):
                data["keyframes"]["LiftHeightKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimProceduralFace):
                data["keyframes"]["ProceduralFaceKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimHeadAngle):
                data["keyframes"]["HeadAngleKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimRobotAudio):
                data["keyframes"]["RobotAudioKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimBackpackLights):
                data["keyframes"]["BackpackLightsKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimFaceAnimation):
                data["keyframes"]["FaceAnimationKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimEvent):
                data["keyframes"]["EventKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimBodyMotion):
                data["keyframes"]["BodyMotionKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimRecordHeading):
                data["keyframes"]["RecordHeadingKeyFrame"].append(keyframe_data)
            elif isinstance(keyframe, AnimTurnToRecordedHeading):
                data["keyframes"]["TurnToRecordedHeadingKeyFrame"].append(keyframe_data)
            else:
                raise RuntimeError("Unexpected keyframe type.")
        return data

    @classmethod
    def from_dict(cls, data: dict):
        keyframes = []
        for keyframe_data in data["keyframes"].get("LiftHeightKeyFrame", []):
            keyframe = AnimLiftHeight.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("ProceduralFaceKeyFrame", []):
            keyframe = AnimProceduralFace.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("HeadAngleKeyFrame", []):
            keyframe = AnimHeadAngle.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("RobotAudioKeyFrame", []):
            keyframe = AnimRobotAudio.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("BackpackLightsKeyFrame", []):
            keyframe = AnimBackpackLights.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("FaceAnimationKeyFrame", []):
            keyframe = AnimFaceAnimation.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("EventKeyFrame", []):
            keyframe = AnimEvent.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("BodyMotionKeyFrame", []):
            keyframe = AnimBodyMotion.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("RecordHeadingKeyFrame", []):
            keyframe = AnimRecordHeading.from_dict(keyframe_data)
            keyframes.append(keyframe)
        for keyframe_data in data["keyframes"].get("TurnToRecordedHeadingKeyFrame", []):
            keyframe = AnimTurnToRecordedHeading.from_dict(keyframe_data)
            keyframes.append(keyframe)
        clip = cls(name=str(data["Name"]), keyframes=keyframes)
        return clip

    def to_fb(self, builder: flatbuffers.Builder):

        head_angle_arr = []
        lift_height_arr = []
        record_heading_arr = []
        turn_to_recorded_heading_arr = []
        body_motion_arr = []
        backpack_lights_arr = []
        face_animation_arr = []
        procedural_face_arr = []
        robot_audio_arr = []
        event_arr = []

        for keyframe in self.keyframes:
            fbkf = keyframe.to_fb(builder)
            if isinstance(keyframe, AnimHeadAngle):
                head_angle_arr.append(fbkf)
            elif isinstance(keyframe, AnimLiftHeight):
                lift_height_arr.append(fbkf)
            elif isinstance(keyframe, AnimRecordHeading):
                record_heading_arr.append(fbkf)
            elif isinstance(keyframe, AnimTurnToRecordedHeading):
                turn_to_recorded_heading_arr.append(fbkf)
            elif isinstance(keyframe, AnimBodyMotion):
                body_motion_arr.append(fbkf)
            elif isinstance(keyframe, AnimBackpackLights):
                backpack_lights_arr.append(fbkf)
            elif isinstance(keyframe, AnimFaceAnimation):
                face_animation_arr.append(fbkf)
            elif isinstance(keyframe, AnimProceduralFace):
                procedural_face_arr.append(fbkf)
            elif isinstance(keyframe, AnimRobotAudio):
                robot_audio_arr.append(fbkf)
            elif isinstance(keyframe, AnimEvent):
                event_arr.append(fbkf)
            else:
                raise RuntimeError("Unexpected keyframe type.")

        CozmoAnim.Keyframes.KeyframesStartHeadAngleKeyFrameVector(builder, len(head_angle_arr))
        for i in reversed(range(len(head_angle_arr))):
            builder.PrependUOffsetTRelative(head_angle_arr[i])
        head_angle_vector = builder.EndVector(len(head_angle_arr))

        CozmoAnim.Keyframes.KeyframesStartLiftHeightKeyFrameVector(builder, len(lift_height_arr))
        for i in reversed(range(len(lift_height_arr))):
            builder.PrependUOffsetTRelative(lift_height_arr[i])
        lift_height_vector = builder.EndVector(len(lift_height_arr))

        CozmoAnim.Keyframes.KeyframesStartRecordHeadingKeyFrameVector(builder, len(record_heading_arr))
        for i in reversed(range(len(record_heading_arr))):
            builder.PrependUOffsetTRelative(record_heading_arr[i])
        record_heading_vector = builder.EndVector(len(record_heading_arr))

        CozmoAnim.Keyframes.KeyframesStartTurnToRecordedHeadingKeyFrameVector(builder,
                                                                              len(turn_to_recorded_heading_arr))
        for i in reversed(range(len(turn_to_recorded_heading_arr))):
            builder.PrependUOffsetTRelative(turn_to_recorded_heading_arr[i])
        turn_to_recorded_heading_vector = builder.EndVector(len(turn_to_recorded_heading_arr))

        CozmoAnim.Keyframes.KeyframesStartBackpackLightsKeyFrameVector(builder, len(backpack_lights_arr))
        for i in reversed(range(len(backpack_lights_arr))):
            builder.PrependUOffsetTRelative(backpack_lights_arr[i])
        backpack_lights_vector = builder.EndVector(len(backpack_lights_arr))

        CozmoAnim.Keyframes.KeyframesStartBodyMotionKeyFrameVector(builder, len(body_motion_arr))
        for i in reversed(range(len(body_motion_arr))):
            builder.PrependUOffsetTRelative(body_motion_arr[i])
        body_motion_vector = builder.EndVector(len(body_motion_arr))

        CozmoAnim.Keyframes.KeyframesStartFaceAnimationKeyFrameVector(builder, len(face_animation_arr))
        for i in reversed(range(len(face_animation_arr))):
            builder.PrependUOffsetTRelative(face_animation_arr[i])
        face_animation_vector = builder.EndVector(len(face_animation_arr))

        CozmoAnim.Keyframes.KeyframesStartProceduralFaceKeyFrameVector(builder, len(procedural_face_arr))
        for i in reversed(range(len(procedural_face_arr))):
            builder.PrependUOffsetTRelative(procedural_face_arr[i])
        procedural_face_vector = builder.EndVector(len(procedural_face_arr))

        CozmoAnim.Keyframes.KeyframesStartRobotAudioKeyFrameVector(builder, len(robot_audio_arr))
        for i in reversed(range(len(robot_audio_arr))):
            builder.PrependUOffsetTRelative(robot_audio_arr[i])
        robot_audio_vector = builder.EndVector(len(robot_audio_arr))

        CozmoAnim.Keyframes.KeyframesStartEventKeyFrameVector(builder, len(event_arr))
        for i in reversed(range(len(event_arr))):
            builder.PrependUOffsetTRelative(event_arr[i])
        event_vector = builder.EndVector(len(event_arr))

        CozmoAnim.Keyframes.KeyframesStart(builder)
        CozmoAnim.Keyframes.KeyframesAddHeadAngleKeyFrame(builder, head_angle_vector)
        CozmoAnim.Keyframes.KeyframesAddLiftHeightKeyFrame(builder, lift_height_vector)
        CozmoAnim.Keyframes.KeyframesAddRecordHeadingKeyFrame(builder, record_heading_vector)
        CozmoAnim.Keyframes.KeyframesAddTurnToRecordedHeadingKeyFrame(builder, turn_to_recorded_heading_vector)
        CozmoAnim.Keyframes.KeyframesAddBodyMotionKeyFrame(builder, body_motion_vector)
        CozmoAnim.Keyframes.KeyframesAddBackpackLightsKeyFrame(builder, backpack_lights_vector)
        CozmoAnim.Keyframes.KeyframesAddFaceAnimationKeyFrame(builder, face_animation_vector)
        CozmoAnim.Keyframes.KeyframesAddProceduralFaceKeyFrame(builder, procedural_face_vector)
        CozmoAnim.Keyframes.KeyframesAddRobotAudioKeyFrame(builder, robot_audio_vector)
        CozmoAnim.Keyframes.KeyframesAddEventKeyFrame(builder, event_vector)
        kfs = CozmoAnim.Keyframes.KeyframesEnd(builder)

        name_str = builder.CreateString(self.name)

        CozmoAnim.AnimClip.AnimClipStart(builder)
        CozmoAnim.AnimClip.AnimClipAddName(builder, name_str)
        CozmoAnim.AnimClip.AnimClipAddKeyframes(builder, kfs)
        fbclip = CozmoAnim.AnimClip.AnimClipEnd(builder)

        return fbclip

    @classmethod
    def from_fb(cls, fbclip: CozmoAnim.AnimClip.AnimClip):
        keyframes = []
        fbkfs = fbclip.Keyframes()

        for i in range(fbkfs.HeadAngleKeyFrameLength()):
            fbkf = fbkfs.HeadAngleKeyFrame(i)
            keyframe = AnimHeadAngle.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.LiftHeightKeyFrameLength()):
            fbkf = fbkfs.LiftHeightKeyFrame(i)
            keyframe = AnimLiftHeight.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.RecordHeadingKeyFrameLength()):
            fbkf = fbkfs.RecordHeadingKeyFrame(i)
            keyframe = AnimRecordHeading.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.TurnToRecordedHeadingKeyFrameLength()):
            fbkf = fbkfs.TurnToRecordedHeadingKeyFrame(i)
            keyframe = AnimTurnToRecordedHeading.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.BodyMotionKeyFrameLength()):
            fbkf = fbkfs.BodyMotionKeyFrame(i)
            keyframe = AnimBodyMotion.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.BackpackLightsKeyFrameLength()):
            fbkf = fbkfs.BackpackLightsKeyFrame(i)
            keyframe = AnimBackpackLights.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.FaceAnimationKeyFrameLength()):
            fbkf = fbkfs.FaceAnimationKeyFrame(i)
            keyframe = AnimFaceAnimation.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.ProceduralFaceKeyFrameLength()):
            fbkf = fbkfs.ProceduralFaceKeyFrame(i)
            keyframe = AnimProceduralFace.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.RobotAudioKeyFrameLength()):
            fbkf = fbkfs.RobotAudioKeyFrame(i)
            keyframe = AnimRobotAudio.from_fb(fbkf)
            keyframes.append(keyframe)

        for i in range(fbkfs.EventKeyFrameLength()):
            fbkf = fbkfs.EventKeyFrame(i)
            keyframe = AnimEvent.from_fb(fbkf)
            keyframes.append(keyframe)

        clip = cls(name=fbclip.Name().decode("utf-8"), keyframes=keyframes)
        return clip


class AnimClips(AnimBase):
    """ Animation clips class. """

    def __init__(self, clips: List[AnimClip] = ()):
        super().__init__()
        self.clips = list(clips)

    def to_dict(self) -> dict:
        data = {
            "clips": [],
        }
        for clip in self.clips:
            data["clips"].append(clip.to_dict())
        return data

    @classmethod
    def from_dict(cls, data: dict):
        clip_list = []
        for clip_dict in data["clips"]:
            clip = AnimClip.from_dict(clip_dict)
            clip_list.append(clip)
        clips = cls(clips=clip_list)
        return clips

    def to_fb(self, builder: flatbuffers.Builder):
        clips_arr = []
        for clip in self.clips:
            fbclip = clip.to_fb(builder)
            clips_arr.append(fbclip)

        CozmoAnim.AnimClips.AnimClipsStartClipsVector(builder, len(clips_arr))
        for i in reversed(range(len(clips_arr))):
            builder.PrependUOffsetTRelative(clips_arr[i])
        clips_vector = builder.EndVector(len(clips_arr))

        CozmoAnim.AnimClips.AnimClipsStart(builder)
        CozmoAnim.AnimClips.AnimClipsAddClips(builder, clips_vector)
        fbclips = CozmoAnim.AnimClips.AnimClipsEnd(builder)

        return fbclips

    @classmethod
    def from_fb(cls, fbclips: CozmoAnim.AnimClips.AnimClips):
        clip_list = []
        for i in range(fbclips.ClipsLength()):
            clip = AnimClip.from_fb(fbclips.Clips(i))
            clip_list.append(clip)
        clips = cls(clips=clip_list)
        return clips

    def to_json_stream(self, f: TextIO) -> None:
        data = self.to_dict()
        json.dump(data, f, indent=2, separators=(",", ": "))

    def to_json_file(self, fspec: str) -> None:
        with open(fspec, "w") as f:
            self.to_json_stream(f)

    @classmethod
    def from_json_stream(cls, f: TextIO):
        data = json.load(f)
        clips = cls.from_dict(data)
        return clips

    @classmethod
    def from_json_file(cls, fspec: str):
        with open(fspec) as f:
            return cls.from_json_stream(f)

    def to_fb_stream(self, f: BinaryIO):
        builder = flatbuffers.Builder(1024)
        fbclips = self.to_fb(builder)
        builder.Finish(fbclips)
        buf = builder.Output()
        f.write(buf)

    def to_fb_file(self, fspec: str):
        with open(fspec, "wb") as f:
            self.to_fb_stream(f)

    @classmethod
    def from_fb_stream(cls, f: BinaryIO):
        buf = f.read()
        fbclips = CozmoAnim.AnimClips.AnimClips.GetRootAsAnimClips(buf, 0)
        clips = cls.from_fb(fbclips)
        return clips

    @classmethod
    def from_fb_file(cls, fspec: str):
        with open(fspec, "rb") as f:
            return cls.from_fb_stream(f)


class AnimHeadAngle(AnimKeyframe):
    """ Head angle keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 duration_ms: int = 0,
                 angle_deg: int = 0,
                 variability_deg: int = 0):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.duration_ms = int(duration_ms)  # uint32
        self.angle_deg = int(angle_deg)  # int8
        self.variability_deg = int(variability_deg)  # uint8 = 0

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "durationTime_ms": self.duration_ms,
            "angle_deg": self.angle_deg,
            "angleVariability_deg": self.variability_deg,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            duration_ms=data.get("durationTime_ms", 0),
            angle_deg=data.get("angle_deg", 0),
            variability_deg=data.get("angleVariability_deg", 0)
        )

    def to_fb(self, builder: flatbuffers.Builder):
        CozmoAnim.HeadAngle.HeadAngleStart(builder)
        CozmoAnim.HeadAngle.HeadAngleAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.HeadAngle.HeadAngleAddDurationTimeMs(builder, self.duration_ms)
        CozmoAnim.HeadAngle.HeadAngleAddAngleDeg(builder, self.angle_deg)
        CozmoAnim.HeadAngle.HeadAngleAddAngleVariabilityDeg(builder, self.variability_deg)
        fbkf = CozmoAnim.HeadAngle.HeadAngleEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.HeadAngle.HeadAngle):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            duration_ms=fbkf.DurationTimeMs(),
            angle_deg=fbkf.AngleDeg(),
            variability_deg=fbkf.AngleVariabilityDeg()
        )


class AnimLiftHeight(AnimKeyframe):
    """ Lift height keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 duration_ms: int = 0,
                 height_mm: int = 0,
                 variability_mm: int = 0):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.duration_ms = int(duration_ms)  # uint32
        self.height_mm = int(height_mm)  # uint8
        self.variability_mm = int(variability_mm)  # uint8 = 0

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "durationTime_ms": self.duration_ms,
            "height_mm": self.height_mm,
            "heightVariability_mm": self.variability_mm,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            duration_ms=data.get("durationTime_ms", 0),
            height_mm=data.get("height_mm", 0),
            variability_mm=data.get("heightVariability_mm"),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        CozmoAnim.LiftHeight.LiftHeightStart(builder)
        CozmoAnim.LiftHeight.LiftHeightAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.LiftHeight.LiftHeightAddDurationTimeMs(builder, self.duration_ms)
        CozmoAnim.LiftHeight.LiftHeightAddHeightMm(builder, self.height_mm)
        CozmoAnim.LiftHeight.LiftHeightAddHeightVariabilityMm(builder, self.variability_mm)
        fbkf = CozmoAnim.LiftHeight.LiftHeightEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.LiftHeight.LiftHeight):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            duration_ms=fbkf.DurationTimeMs(),
            height_mm=fbkf.HeightMm(),
            variability_mm=fbkf.HeightVariabilityMm()
        )


class AnimRecordHeading(AnimKeyframe):
    """ Record heading keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        CozmoAnim.RecordHeading.RecordHeadingStart(builder)
        CozmoAnim.RecordHeading.RecordHeadingAddTriggerTimeMs(builder, self.trigger_time_ms)
        fbkf = CozmoAnim.RecordHeading.RecordHeadingEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.RecordHeading.RecordHeading):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs()
        )


class AnimTurnToRecordedHeading(AnimKeyframe):
    """ Turn-to-recorded-heading keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 duration_ms: int = 0,
                 offset_deg: int = 0,
                 speed_deg_per_sec: int = 0,
                 accel_deg_per_sec_2: int = 1000,
                 decel_deg_per_sec_2: int = 1000,
                 tolerance_deg: int = 2,
                 num_half_revs: int = 0,
                 use_shortest_dir: bool = False):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.duration_ms = int(duration_ms)  # uint32
        self.offset_deg = int(offset_deg)  # int16 = 0
        self.speed_deg_per_sec = int(speed_deg_per_sec)  # int16
        self.accel_deg_per_sec_2 = int(accel_deg_per_sec_2)  # int16 = 1000
        self.decel_deg_per_sec_2 = int(decel_deg_per_sec_2)  # int16 = 1000
        self.tolerance_deg = int(tolerance_deg)  # uint16 = 2
        self.num_half_revs = int(num_half_revs)  # uint16 = 0
        self.use_shortest_dir = bool(use_shortest_dir)  # boot = false

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "durationTime_ms": self.duration_ms,
            "offset_deg": self.offset_deg,
            "speed_degPerSec": self.speed_deg_per_sec,
            "accel_degPerSec2": self.accel_deg_per_sec_2,
            "decel_degPerSec2": self.decel_deg_per_sec_2,
            "tolerance_deg": self.tolerance_deg,
            "numHalfRevs": self.num_half_revs,
            "useShortestDir": self.use_shortest_dir,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            duration_ms=data.get("durationTime_ms", 0),
            offset_deg=data.get("offset_deg", 0),
            speed_deg_per_sec=data.get("speed_degPerSec", 0),
            accel_deg_per_sec_2=data.get("accel_degPerSec2", 1000),
            decel_deg_per_sec_2=data.get("decel_degPerSec2", 1000),
            tolerance_deg=data.get("tolerance_deg", 2),
            num_half_revs=data.get("numHalfRevs", 0),
            use_shortest_dir=data.get("useShortestDir", False),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingStart(builder)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddDurationTimeMs(builder, self.duration_ms)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddOffsetDeg(builder, self.offset_deg)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddSpeedDegPerSec(builder, self.speed_deg_per_sec)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddAccelDegPerSec2(builder, self.accel_deg_per_sec_2)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddDecelDegPerSec2(builder, self.decel_deg_per_sec_2)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddToleranceDeg(builder, self.tolerance_deg)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddNumHalfRevs(builder, self.num_half_revs)
        CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingAddUseShortestDir(builder, self.use_shortest_dir)
        fbkf = CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeadingEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.TurnToRecordedHeading.TurnToRecordedHeading):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            duration_ms=fbkf.DurationTimeMs(),
            offset_deg=fbkf.OffsetDeg(),
            speed_deg_per_sec=fbkf.SpeedDegPerSec(),
            accel_deg_per_sec_2=fbkf.AccelDegPerSec2(),
            decel_deg_per_sec_2=fbkf.DecelDegPerSec2(),
            tolerance_deg=fbkf.ToleranceDeg(),
            num_half_revs=fbkf.NumHalfRevs(),
            use_shortest_dir=fbkf.UseShortestDir()
        )


class AnimBodyMotion(AnimKeyframe):
    """ Body motion keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 duration_ms: int = 0,
                 radius_mm: Union[float, str] = "STRAIGHT",
                 speed: int = 0):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.duration_ms = int(duration_ms)  # uint32
        try:
            self.radius_mm = float(radius_mm)
        except ValueError:
            self.radius_mm = str(radius_mm)  # string (required)
        self.speed = int(speed)  # int16

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "durationTime_ms": self.duration_ms,
            "radius_mm": str(self.radius_mm),
            "speed": self.speed,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            duration_ms=data.get("durationTime_ms", 0),
            radius_mm=data.get("radius_mm", "STRAIGHT"),
            speed=data.get("speed", 0),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        radius_mm_str = builder.CreateString(str(self.radius_mm))
        CozmoAnim.BodyMotion.BodyMotionStart(builder)
        CozmoAnim.BodyMotion.BodyMotionAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.BodyMotion.BodyMotionAddDurationTimeMs(builder, self.duration_ms)
        CozmoAnim.BodyMotion.BodyMotionAddRadiusMm(builder, radius_mm_str)
        CozmoAnim.BodyMotion.BodyMotionAddSpeed(builder, self.speed)
        fbkf = CozmoAnim.BodyMotion.BodyMotionEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.BodyMotion.BodyMotion):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            duration_ms=fbkf.DurationTimeMs(),
            radius_mm=fbkf.RadiusMm().decode("utf-8"),
            speed=fbkf.Speed()
        )


class AnimLight(object):
    """ Light color class. """

    def __init__(self, red: int = 0, green: int = 0, blue: int = 0, ir: int = 0):
        self.red = int(red)
        self.green = int(green)
        self.blue = int(blue)
        self.ir = int(ir)

    def to_dict(self) -> list:
        return [
            self.red,
            self.green,
            self.blue,
            self.ir,
        ]

    @classmethod
    def from_dict(cls, data):
        return cls(
            red=data[0],
            green=data[1],
            blue=data[2],
            ir=data[3]
        )


class AnimBackpackLights(AnimKeyframe):
    """ Backpack lights keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 duration_ms: int = 0,
                 left: AnimLight = AnimLight(),
                 front: AnimLight = AnimLight(),
                 middle: AnimLight = AnimLight(),
                 back: AnimLight = AnimLight(),
                 right: AnimLight = AnimLight()):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.duration_ms = int(duration_ms)  # uint32
        self.left = left  # [float]
        self.front = front  # [float]
        self.middle = middle  # [float]
        self.back = back  # [float]
        self.right = right  # [float]

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "durationTime_ms": self.duration_ms,
            "Left": self.left.to_dict(),
            "Front": self.front.to_dict(),
            "Middle": self.middle.to_dict(),
            "Back": self.back.to_dict(),
            "Right": self.right.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            duration_ms=data.get("durationTime_ms", 0),
            left=AnimLight.from_dict(data.get("Left", [0, 0, 0, 0])),
            front=AnimLight.from_dict(data.get("Front", [0, 0, 0, 0])),
            middle=AnimLight.from_dict(data.get("Middle", [0, 0, 0, 0])),
            back=AnimLight.from_dict(data.get("Back", [0, 0, 0, 0])),
            right=AnimLight.from_dict(data.get("Right", [0, 0, 0, 0])),
        )

    def to_fb(self, builder: flatbuffers.Builder):

        led_vectors = []
        for led in (self.left, self.front, self.middle, self.back, self.right):
            CozmoAnim.BackpackLights.BackpackLightsStartLeftVector(builder, 4)
            builder.PrependFloat32(led.ir)
            builder.PrependFloat32(led.blue)
            builder.PrependFloat32(led.green)
            builder.PrependFloat32(led.red)
            led_vectors.append(builder.EndVector(4))

        CozmoAnim.BackpackLights.BackpackLightsStart(builder)
        CozmoAnim.BackpackLights.BackpackLightsAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.BackpackLights.BackpackLightsAddDurationTimeMs(builder, self.duration_ms)
        CozmoAnim.BackpackLights.BackpackLightsAddLeft(builder, led_vectors[0])
        CozmoAnim.BackpackLights.BackpackLightsAddFront(builder, led_vectors[1])
        CozmoAnim.BackpackLights.BackpackLightsAddMiddle(builder, led_vectors[2])
        CozmoAnim.BackpackLights.BackpackLightsAddBack(builder, led_vectors[3])
        CozmoAnim.BackpackLights.BackpackLightsAddRight(builder, led_vectors[4])
        fbkf = CozmoAnim.BackpackLights.BackpackLightsEnd(builder)

        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.BackpackLights.BackpackLights):
        assert fbkf.LeftLength() == 4
        assert fbkf.FrontLength() == 4
        assert fbkf.MiddleLength() == 4
        assert fbkf.BackLength() == 4
        assert fbkf.RightLength() == 4
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            duration_ms=fbkf.DurationTimeMs(),
            left=AnimLight(fbkf.Left(0), fbkf.Left(1), fbkf.Left(2), fbkf.Left(3)),
            front=AnimLight(fbkf.Front(0), fbkf.Front(1), fbkf.Front(2), fbkf.Front(3)),
            middle=AnimLight(fbkf.Middle(0), fbkf.Middle(1), fbkf.Middle(2), fbkf.Middle(3)),
            back=AnimLight(fbkf.Back(0), fbkf.Back(1), fbkf.Back(2), fbkf.Back(3)),
            right=AnimLight(fbkf.Right(0), fbkf.Right(1), fbkf.Right(2), fbkf.Right(3))
        )


class AnimFaceAnimation(AnimKeyframe):
    """ Face animation keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 anim_name: str = ""):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.anim_name = str(anim_name)

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "animName": self.anim_name,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            anim_name=data.get("animName", ""),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        anim_name_str = builder.CreateString(self.anim_name)
        CozmoAnim.FaceAnimation.FaceAnimationStart(builder)
        CozmoAnim.FaceAnimation.FaceAnimationAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.FaceAnimation.FaceAnimationAddAnimName(builder, anim_name_str)
        fbkf = CozmoAnim.FaceAnimation.FaceAnimationEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.FaceAnimation.FaceAnimation):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            anim_name=fbkf.AnimName().decode("utf-8")
        )


class AnimProceduralFace(AnimKeyframe):
    """ Procedural face keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 angle: float = 0.0,
                 center_x: float = 0.0,
                 center_y: float = 0.0,
                 scale_x: float = 1.0,
                 scale_y: float = 1.0,
                 left_eye: List[float] = (),
                 right_eye: List[float] = ()):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.angle = float(angle)  # float = 0.0
        self.center_x = float(center_x)  # float = 0.0
        self.center_y = float(center_y)  # float = 0.0
        self.scale_x = float(scale_x)  # float = 1.0
        self.scale_y = float(scale_y)  # float = 1.0
        assert(len(left_eye) == 19)
        self.left_eye = list(left_eye)  # [float]
        assert (len(right_eye) == 19)
        self.right_eye = list(right_eye)  # [float]

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "faceAngle": self.angle,
            "faceCenterX": self.center_x,
            "faceCenterY": self.center_y,
            "faceScaleX": self.scale_x,
            "faceScaleY": self.scale_y,
            "leftEye": list(self.left_eye),
            "rightEye": list(self.right_eye),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            angle=data.get("faceAngle", 0.0),
            center_x=data.get("faceCenterX", 0.0),
            center_y=data.get("faceCenterY", 0.0),
            scale_x=data.get("faceScaleX", 1.0),
            scale_y=data.get("faceScaleY", 1.0),
            left_eye=data.get("leftEye", ()),
            right_eye=data.get("rightEye", ()),
        )

    def to_fb(self, builder: flatbuffers.Builder):

        eye_vectors = []
        for eye in (self.left_eye, self.right_eye):
            CozmoAnim.ProceduralFace.ProceduralFaceStartLeftEyeVector(builder, 19)
            for i in reversed(range(19)):
                builder.PrependFloat32(eye[i])
            eye_vectors.append(builder.EndVector(19))

        CozmoAnim.ProceduralFace.ProceduralFaceStart(builder)
        CozmoAnim.ProceduralFace.ProceduralFaceAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.ProceduralFace.ProceduralFaceAddFaceAngle(builder, self.angle)
        CozmoAnim.ProceduralFace.ProceduralFaceAddFaceCenterX(builder, self.center_x)
        CozmoAnim.ProceduralFace.ProceduralFaceAddFaceCenterY(builder, self.center_y)
        CozmoAnim.ProceduralFace.ProceduralFaceAddFaceScaleX(builder, self.scale_x)
        CozmoAnim.ProceduralFace.ProceduralFaceAddFaceScaleY(builder, self.scale_y)
        CozmoAnim.ProceduralFace.ProceduralFaceAddLeftEye(builder, eye_vectors[0])
        CozmoAnim.ProceduralFace.ProceduralFaceAddRightEye(builder, eye_vectors[1])
        fbkf = CozmoAnim.ProceduralFace.ProceduralFaceEnd(builder)

        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.ProceduralFace.ProceduralFace):
        assert fbkf.LeftEyeLength() == 19
        left_eye = [fbkf.LeftEye(i) for i in range(fbkf.LeftEyeLength())]
        assert fbkf.RightEyeLength() == 19
        right_eye = [fbkf.RightEye(i) for i in range(fbkf.RightEyeLength())]
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            angle=fbkf.FaceAngle(),
            center_x=fbkf.FaceCenterX(),
            center_y=fbkf.FaceCenterY(),
            scale_x=fbkf.FaceScaleX(),
            scale_y=fbkf.FaceScaleY(),
            left_eye=left_eye,
            right_eye=right_eye
        )


class AnimRobotAudio(AnimKeyframe):
    """ Robot audio keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 audio_event_ids: List[int] = (),
                 volume: float = 1.0,
                 probabilities: List[float] = (),
                 has_alts: bool = True):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.audio_event_ids = list(audio_event_ids)  # [int32]
        self.volume = float(volume)  # float = 1.0
        self.probabilities = list(probabilities) if isinstance(probabilities, list) else [probabilities]  # [float]
        self.has_alts = bool(has_alts)  # bool = true

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "audioEventId": list(self.audio_event_ids),
            "volume": self.volume,
            "probability": list(self.probabilities),
            "hasAlts": self.has_alts,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            audio_event_ids=data.get("audioEventId", ()),
            volume=data.get("volume", 1.0),
            probabilities=data.get("probability", ()),
            has_alts=data.get("hasAlts", True),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        CozmoAnim.RobotAudio.RobotAudioStartAudioEventIdVector(builder, len(self.audio_event_ids))
        for i in reversed(range(len(self.audio_event_ids))):
            builder.PrependInt64(self.audio_event_ids[i])
        audio_event_id_vector = builder.EndVector(len(self.audio_event_ids))

        CozmoAnim.RobotAudio.RobotAudioStartProbabilityVector(builder, len(self.probabilities))
        for i in reversed(range(len(self.probabilities))):
            builder.PrependFloat32(self.probabilities[i])
        probability_vector = builder.EndVector(len(self.probabilities))

        CozmoAnim.RobotAudio.RobotAudioStart(builder)
        CozmoAnim.RobotAudio.RobotAudioAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.RobotAudio.RobotAudioAddAudioEventId(builder, audio_event_id_vector)
        CozmoAnim.RobotAudio.RobotAudioAddVolume(builder, self.volume)
        CozmoAnim.RobotAudio.RobotAudioAddProbability(builder, probability_vector)
        CozmoAnim.RobotAudio.RobotAudioAddHasAlts(builder, self.has_alts)
        fbkf = CozmoAnim.RobotAudio.RobotAudioEnd(builder)

        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.RobotAudio.RobotAudio):
        audio_event_ids = []
        for j in range(fbkf.AudioEventIdLength()):
            audio_event_ids.append(fbkf.AudioEventId(j))
        probabilities = []
        for j in range(fbkf.ProbabilityLength()):
            probabilities.append(fbkf.Probability(j))
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            audio_event_ids=audio_event_ids,
            volume=fbkf.Volume(),
            probabilities=probabilities,
            has_alts=fbkf.HasAlts()
        )


class AnimEvent(AnimKeyframe):
    """ Event keyframe class. """

    def __init__(self,
                 trigger_time_ms: int = 0,
                 event_id: str = ""):
        super().__init__()
        self.trigger_time_ms = int(trigger_time_ms)  # uint32
        self.event_id = str(event_id)

    def to_dict(self) -> dict:
        return {
            "triggerTime_ms": self.trigger_time_ms,
            "event_id": self.event_id,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            trigger_time_ms=data.get("triggerTime_ms", 0),
            event_id=data.get("event_id", ""),
        )

    def to_fb(self, builder: flatbuffers.Builder):
        event_id_str = builder.CreateString(self.event_id)
        CozmoAnim.Event.EventStart(builder)
        CozmoAnim.Event.EventAddTriggerTimeMs(builder, self.trigger_time_ms)
        CozmoAnim.Event.EventAddEventId(builder, event_id_str)
        fbkf = CozmoAnim.Event.EventEnd(builder)
        return fbkf

    @classmethod
    def from_fb(cls, fbkf: CozmoAnim.Event.Event):
        return cls(
            trigger_time_ms=fbkf.TriggerTimeMs(),
            event_id=fbkf.EventId().decode("utf-8")
        )


class ClipMetadata(object):
    """ Animation clip metadata class. """
    def __init__(self,
                 fspec: str,
                 index: int,
                 name: str,
                 has_head_angle_track: bool,
                 has_lift_height_track: bool,
                 has_record_heading_track: bool,
                 has_turn_to_recorded_heading_track: bool,
                 has_body_motion_track: bool,
                 has_backpack_lights_track: bool,
                 has_face_animation_track: bool,
                 has_procedural_face_track: bool,
                 has_robot_audio_track: bool,
                 has_event_track: bool):
        self.fspec = str(fspec)
        self.index = int(index)
        self.name = str(name)
        self.has_head_angle_track = bool(has_head_angle_track)
        self.has_lift_height_track = bool(has_lift_height_track)
        self.has_record_heading_track = bool(has_record_heading_track)
        self.has_turn_to_recorded_heading_track = bool(has_turn_to_recorded_heading_track)
        self.has_body_motion_track = bool(has_body_motion_track)
        self.has_backpack_lights_track = bool(has_backpack_lights_track)
        self.has_face_animation_track = bool(has_face_animation_track)
        self.has_procedural_face_track = bool(has_procedural_face_track)
        self.has_robot_audio_track = bool(has_robot_audio_track)
        self.has_event_track = bool(has_event_track)


def get_clip_metadata(dspec: str) -> Dict[str, ClipMetadata]:
    """ Retrieve clip metadata from animation FlatBuffers .bin files. """
    res = {}
    # Find all animation files.
    for fspec in glob.glob(os.path.join(dspec, "*.bin")):
        with open(fspec, "rb") as f:
            buf = f.read()
        fbclips = CozmoAnim.AnimClips.AnimClips.GetRootAsAnimClips(buf, 0)
        # Read properties of all clips.
        for i in range(fbclips.ClipsLength()):
            fbclip = fbclips.Clips(i)
            fbkfs = fbclip.Keyframes()
            metadata = ClipMetadata(
                fspec,
                i,
                fbclip.Name().decode("utf-8"),
                fbkfs.HeadAngleKeyFrameLength() > 0,
                fbkfs.LiftHeightKeyFrameLength() > 0,
                fbkfs.RecordHeadingKeyFrameLength() > 0,
                fbkfs.TurnToRecordedHeadingKeyFrameLength() > 0,
                fbkfs.BodyMotionKeyFrameLength() > 0,
                fbkfs.BackpackLightsKeyFrameLength() > 0,
                fbkfs.FaceAnimationKeyFrameLength() > 0,
                fbkfs.ProceduralFaceKeyFrameLength() > 0,
                fbkfs.RobotAudioKeyFrameLength() > 0,
                fbkfs.EventKeyFrameLength() > 0)
            res[metadata.name] = metadata
            if metadata.has_procedural_face_track and \
               not metadata.has_body_motion_track and \
               not metadata.has_lift_height_track and \
               not metadata.has_head_angle_track and \
               not metadata.has_robot_audio_track:
                print("{}:{}:{}".format(metadata.fspec, metadata.index, metadata.name))
    return res




from typing import Optional
import sys
import os
import logging

from . import logger, logger_protocol
from . import client


__all__ = [
    'setup_basic_logging',
    'run_program',
]


def setup_basic_logging(log_level: Optional[str] = None, protocol_log_level: Optional[str] = None,
                        target=sys.stderr) -> None:
    if log_level is None:
        log_level = os.environ.get('PYCOZMO_LOG_LEVEL', logging.INFO)
    if protocol_log_level is None:
        protocol_log_level = os.environ.get('PYCOZMO_PROTOCOL_LOG_LEVEL', logging.WARNING)
    handler = logging.StreamHandler(stream=target)
    formatter = logging.Formatter(
        fmt="%(asctime)s.%(msecs)03d %(name)-20s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(log_level)
    logger_protocol.addHandler(handler)
    logger_protocol.setLevel(protocol_log_level)


def run_program(f: callable, log_level: Optional[str] = None, protocol_log_level: Optional[str] = None,
                protocol_log_messages: Optional[list] = None) -> None:
    setup_basic_logging(log_level=log_level, protocol_log_level=protocol_log_level)

    cli = client.Client(protocol_log_messages=protocol_log_messages)
    cli.start()
    cli.connect()
    cli.wait_for_robot()

    try:
        f(cli)
    except KeyboardInterrupt:
        logger.info("Interrupted...")
    finally:
        cli.disconnect()
        cli.stop()

    logger.info("Done.")




import collections
import threading

from . import exception
from . import robot


__all__ = [
    "Event",
    "Handler",
    "Dispatcher",
]


class Event(object):
    """ Base class for events. """


class Handler(object):
    """ Event handler class. """
    def __init__(self, f: callable, one_shot: bool):
        self.f = f
        self.one_shot = one_shot


class EvtRobotFound(Event):
    """ Triggered when the robot has been first connected. """


class EvtRobotReady(Event):
    """ Triggered when the robot has been initialized and is ready for commands. """


class EvtNewRawCameraImage(Event):
    """ Triggered when a new raw image is received from the robot's camera. """


class EvtRobotMovingChange(Event):
    pass


class EvtRobotCarryingBlockChange(Event):
    pass


class EvtRobotPickingOrPlacingChange(Event):
    pass


class EvtRobotPickedUpChange(Event):
    pass


class EvtRobotBodyAccModeChange(Event):
    pass


class EvtRobotFallingChange(Event):
    pass


class EvtRobotAnimatingChange(Event):
    pass


class EvtRobotPathingChange(Event):
    pass


class EvtRobotLiftInPositionChange(Event):
    pass


class EvtRobotHeadInPositionChange(Event):
    pass


class EvtRobotAnimBufferFullChange(Event):
    pass


class EvtRobotAnimatingIdleChange(Event):
    pass


class EvtRobotOnChargerChange(Event):
    pass


class EvtRobotChargingChange(Event):
    pass


class EvtCliffDetectedChange(Event):
    pass


class EvtRobotWheelsMovingChange(Event):
    pass


class EvtChargerOOSChange(Event):
    pass


STATUS_EVENTS = {
    robot.RobotStatusFlag.IS_MOVING: EvtRobotMovingChange,
    robot.RobotStatusFlag.IS_CARRYING_BLOCK: EvtRobotCarryingBlockChange,
    robot.RobotStatusFlag.IS_PICKING_OR_PLACING: EvtRobotPickingOrPlacingChange,
    robot.RobotStatusFlag.IS_PICKED_UP: EvtRobotPickedUpChange,
    robot.RobotStatusFlag.IS_BODY_ACC_MODE: EvtRobotBodyAccModeChange,
    robot.RobotStatusFlag.IS_FALLING: EvtRobotFallingChange,
    robot.RobotStatusFlag.IS_ANIMATING: EvtRobotAnimatingChange,
    robot.RobotStatusFlag.IS_PATHING: EvtRobotPathingChange,
    robot.RobotStatusFlag.LIFT_IN_POS: EvtRobotLiftInPositionChange,
    robot.RobotStatusFlag.HEAD_IN_POS: EvtRobotHeadInPositionChange,
    robot.RobotStatusFlag.IS_ANIM_BUFFER_FULL: EvtRobotAnimBufferFullChange,
    robot.RobotStatusFlag.IS_ANIMATING_IDLE: EvtRobotAnimatingChange,
    robot.RobotStatusFlag.IS_ON_CHARGER: EvtRobotOnChargerChange,
    robot.RobotStatusFlag.IS_CHARGING: EvtRobotChargingChange,
    robot.RobotStatusFlag.CLIFF_DETECTED: EvtCliffDetectedChange,
    robot.RobotStatusFlag.ARE_WHEELS_MOVING: EvtRobotWheelsMovingChange,
    robot.RobotStatusFlag.IS_CHARGER_OOS: EvtChargerOOSChange,
}


class EvtRobotStateUpdated(Event):
    """ Triggered when a new robot state is received. """


class Dispatcher(object):
    """ Event dispatcher class. """

    def __init__(self):
        super().__init__()
        self.dispatch_handlers = collections.defaultdict(list)

    def add_handler(self, event, f, one_shot=False):
        handler = Handler(f, one_shot=one_shot)
        self.dispatch_handlers[event].append(handler)
        return handler

    def del_handler(self, event, handler):
        for i, _handler in enumerate(self.dispatch_handlers[event]):
            if _handler == handler:
                del self.dispatch_handlers[event][i]
                return

    def del_all_handlers(self):
        self.dispatch_handlers = collections.defaultdict(list)

    def dispatch(self, event, *args, **kwargs):
        handlers = []
        for i, handler in enumerate(self.dispatch_handlers[event]):
            if handler.one_shot:
                # Delete one-shot handlers prior to actual dispatch
                del self.dispatch_handlers[event][i]
            handlers.append(handler)

        for handler in handlers:
            handler.f(*args, **kwargs)

    def wait_for(self, evt, timeout: float = None) -> None:
        e = threading.Event()
        self.add_handler(evt, lambda *args: e.set(), one_shot=True)
        if not e.wait(timeout):
            raise exception.Timeout("Failed to receive event in time.")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class Keyframes(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsKeyframes(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Keyframes()
        x.Init(buf, n + offset)
        return x

    # Keyframes
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Keyframes
    def LiftHeightKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .LiftHeight import LiftHeight
            obj = LiftHeight()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def LiftHeightKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def ProceduralFaceKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .ProceduralFace import ProceduralFace
            obj = ProceduralFace()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def ProceduralFaceKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def HeadAngleKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .HeadAngle import HeadAngle
            obj = HeadAngle()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def HeadAngleKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def RobotAudioKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .RobotAudio import RobotAudio
            obj = RobotAudio()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def RobotAudioKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def BackpackLightsKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .BackpackLights import BackpackLights
            obj = BackpackLights()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def BackpackLightsKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def FaceAnimationKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .FaceAnimation import FaceAnimation
            obj = FaceAnimation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def FaceAnimationKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def EventKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Event import Event
            obj = Event()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def EventKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def BodyMotionKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .BodyMotion import BodyMotion
            obj = BodyMotion()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def BodyMotionKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def RecordHeadingKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .RecordHeading import RecordHeading
            obj = RecordHeading()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def RecordHeadingKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Keyframes
    def TurnToRecordedHeadingKeyFrame(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .TurnToRecordedHeading import TurnToRecordedHeading
            obj = TurnToRecordedHeading()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Keyframes
    def TurnToRecordedHeadingKeyFrameLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def KeyframesStart(builder): builder.StartObject(10)
def KeyframesAddLiftHeightKeyFrame(builder, LiftHeightKeyFrame): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(LiftHeightKeyFrame), 0)
def KeyframesStartLiftHeightKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddProceduralFaceKeyFrame(builder, ProceduralFaceKeyFrame): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(ProceduralFaceKeyFrame), 0)
def KeyframesStartProceduralFaceKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddHeadAngleKeyFrame(builder, HeadAngleKeyFrame): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(HeadAngleKeyFrame), 0)
def KeyframesStartHeadAngleKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddRobotAudioKeyFrame(builder, RobotAudioKeyFrame): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(RobotAudioKeyFrame), 0)
def KeyframesStartRobotAudioKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddBackpackLightsKeyFrame(builder, BackpackLightsKeyFrame): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(BackpackLightsKeyFrame), 0)
def KeyframesStartBackpackLightsKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddFaceAnimationKeyFrame(builder, FaceAnimationKeyFrame): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(FaceAnimationKeyFrame), 0)
def KeyframesStartFaceAnimationKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddEventKeyFrame(builder, EventKeyFrame): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(EventKeyFrame), 0)
def KeyframesStartEventKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddBodyMotionKeyFrame(builder, BodyMotionKeyFrame): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(BodyMotionKeyFrame), 0)
def KeyframesStartBodyMotionKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddRecordHeadingKeyFrame(builder, RecordHeadingKeyFrame): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(RecordHeadingKeyFrame), 0)
def KeyframesStartRecordHeadingKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesAddTurnToRecordedHeadingKeyFrame(builder, TurnToRecordedHeadingKeyFrame): builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(TurnToRecordedHeadingKeyFrame), 0)
def KeyframesStartTurnToRecordedHeadingKeyFrameVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def KeyframesEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class TurnToRecordedHeading(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTurnToRecordedHeading(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TurnToRecordedHeading()
        x.Init(buf, n + offset)
        return x

    # TurnToRecordedHeading
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TurnToRecordedHeading
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # TurnToRecordedHeading
    def DurationTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # TurnToRecordedHeading
    def OffsetDeg(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # TurnToRecordedHeading
    def SpeedDegPerSec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # TurnToRecordedHeading
    def AccelDegPerSec2(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 1000

    # TurnToRecordedHeading
    def DecelDegPerSec2(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 1000

    # TurnToRecordedHeading
    def ToleranceDeg(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 2

    # TurnToRecordedHeading
    def NumHalfRevs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # TurnToRecordedHeading
    def UseShortestDir(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def TurnToRecordedHeadingStart(builder): builder.StartObject(9)
def TurnToRecordedHeadingAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def TurnToRecordedHeadingAddDurationTimeMs(builder, durationTimeMs): builder.PrependUint32Slot(1, durationTimeMs, 0)
def TurnToRecordedHeadingAddOffsetDeg(builder, offsetDeg): builder.PrependInt16Slot(2, offsetDeg, 0)
def TurnToRecordedHeadingAddSpeedDegPerSec(builder, speedDegPerSec): builder.PrependInt16Slot(3, speedDegPerSec, 0)
def TurnToRecordedHeadingAddAccelDegPerSec2(builder, accelDegPerSec2): builder.PrependInt16Slot(4, accelDegPerSec2, 1000)
def TurnToRecordedHeadingAddDecelDegPerSec2(builder, decelDegPerSec2): builder.PrependInt16Slot(5, decelDegPerSec2, 1000)
def TurnToRecordedHeadingAddToleranceDeg(builder, toleranceDeg): builder.PrependUint16Slot(6, toleranceDeg, 2)
def TurnToRecordedHeadingAddNumHalfRevs(builder, numHalfRevs): builder.PrependUint16Slot(7, numHalfRevs, 0)
def TurnToRecordedHeadingAddUseShortestDir(builder, useShortestDir): builder.PrependBoolSlot(8, useShortestDir, 0)
def TurnToRecordedHeadingEnd(builder): return builder.EndObject()




import json
import io
import unittest

from pycozmo.anim_encoder import AnimClips


class TestClips(unittest.TestCase):

    def test_no_clips(self):
        data = json.loads("""
{
  "clips": []
}       
        """)
        # Load from JSON
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips), 0)

        # Save to FB
        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        # Load from FB
        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips), 0)

        # Save to JSON
        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestClip(unittest.TestCase):

    def test_empty_clip(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips), 1)
        self.assertEqual(clips.clips[0].name, "test")
        self.assertEqual(len(clips.clips[0].keyframes), 0)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips), 1)
        self.assertEqual(clips.clips[0].name, "test")
        self.assertEqual(len(clips.clips[0].keyframes), 0)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestHeadAngle(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [
          {
            "triggerTime_ms": 1,
            "durationTime_ms": 2,
            "angle_deg": 3,
            "angleVariability_deg": 4
          }
        ],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.angle_deg, 3)
        self.assertEqual(keyframe.variability_deg, 4)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.angle_deg, 3)
        self.assertEqual(keyframe.variability_deg, 4)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestLiftHeight(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [
          {
            "triggerTime_ms": 1,
            "durationTime_ms": 2,
            "height_mm": 3,
            "heightVariability_mm": 4
          }
        ],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.height_mm, 3)
        self.assertEqual(keyframe.variability_mm, 4)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.height_mm, 3)
        self.assertEqual(keyframe.variability_mm, 4)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestRecordHeading(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [
          {
            "triggerTime_ms": 1
          }
        ],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestBodyMotion(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [
          {
            "triggerTime_ms": 1,
            "durationTime_ms": 2,
            "radius_mm": "3.0",
            "speed": 4
          }
        ],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.radius_mm, 3)
        self.assertEqual(keyframe.speed, 4)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.radius_mm, 3)
        self.assertEqual(keyframe.speed, 4)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestBackpackLights(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [
          {
            "triggerTime_ms": 1,
            "durationTime_ms": 2,
            "Left": [
              1.0,
              2.0,
              3.0,
              4.0
            ],
            "Right": [
              1.0,
              2.0,
              3.0,
              4.0
            ],
            "Front": [
              1.0,
              2.0,
              3.0,
              4.0
            ],
            "Middle": [
              1.0,
              2.0,
              3.0,
              4.0
            ],
            "Back": [
              1.0,
              2.0,
              3.0,
              4.0
            ]
          }
        ],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        for led in (keyframe.left, keyframe.front, keyframe.middle, keyframe.back, keyframe.right):
            self.assertEqual(led.red, 1)
            self.assertEqual(led.green, 2)
            self.assertEqual(led.blue, 3)
            self.assertEqual(led.ir, 4)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        for led in (keyframe.left, keyframe.front, keyframe.middle, keyframe.back, keyframe.right):
            self.assertEqual(led.red, 1)
            self.assertEqual(led.green, 2)
            self.assertEqual(led.blue, 3)
            self.assertEqual(led.ir, 4)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestTurnToRecordedHeading(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": [
          {
            "triggerTime_ms": 1,
            "durationTime_ms": 2,
            "offset_deg": 3,
            "speed_degPerSec": 4,
            "accel_degPerSec2": 5,
            "decel_degPerSec2": 6,
            "tolerance_deg": 7,
            "numHalfRevs": 8,
            "useShortestDir": true
          }
        ]
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.offset_deg, 3)
        self.assertEqual(keyframe.speed_deg_per_sec, 4)
        self.assertEqual(keyframe.accel_deg_per_sec_2, 5)
        self.assertEqual(keyframe.decel_deg_per_sec_2, 6)
        self.assertEqual(keyframe.tolerance_deg, 7)
        self.assertEqual(keyframe.num_half_revs, 8)
        self.assertTrue(keyframe.use_shortest_dir)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.duration_ms, 2)
        self.assertEqual(keyframe.offset_deg, 3)
        self.assertEqual(keyframe.speed_deg_per_sec, 4)
        self.assertEqual(keyframe.accel_deg_per_sec_2, 5)
        self.assertEqual(keyframe.decel_deg_per_sec_2, 6)
        self.assertEqual(keyframe.tolerance_deg, 7)
        self.assertEqual(keyframe.num_half_revs, 8)
        self.assertTrue(keyframe.use_shortest_dir)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestFaceAnimation(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [
          {
            "triggerTime_ms": 1,
            "animName": "test"
          }
        ],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.anim_name, "test")

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.anim_name, "test")

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestProceduralFace(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [
          {
            "triggerTime_ms": 1,
            "faceAngle": 2.0,
            "faceCenterX": 3.0,
            "faceCenterY": 4.0,
            "faceScaleX": 5.0,
            "faceScaleY": 6.0,
            "leftEye": [
              1.0,
              2.0,
              3.0,
              4.0,
              5.0,
              6.0,
              7.0,
              8.0,
              9.0,
              10.0,
              11.0,
              12.0,
              13.0,
              14.0,
              15.0,
              16.0,
              17.0,
              18.0,
              19.0
            ],
            "rightEye": [
              1.0,
              2.0,
              3.0,
              4.0,
              5.0,
              6.0,
              7.0,
              8.0,
              9.0,
              10.0,
              11.0,
              12.0,
              13.0,
              14.0,
              15.0,
              16.0,
              17.0,
              18.0,
              19.0
            ]
          }
        ],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.angle, 2.0)
        self.assertEqual(keyframe.center_x, 3.0)
        self.assertEqual(keyframe.center_y, 4.0)
        self.assertEqual(keyframe.scale_x, 5.0)
        self.assertEqual(keyframe.scale_y, 6.0)
        for eye in (keyframe.left_eye, keyframe.right_eye):
            for i in range(19):
                self.assertEqual(eye[i], float(i + 1))

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.angle, 2.0)
        self.assertEqual(keyframe.center_x, 3.0)
        self.assertEqual(keyframe.center_y, 4.0)
        self.assertEqual(keyframe.scale_x, 5.0)
        self.assertEqual(keyframe.scale_y, 6.0)
        for eye in (keyframe.left_eye, keyframe.right_eye):
            for i in range(19):
                self.assertEqual(eye[i], float(i + 1))

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestRobotAudio(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [
          {
            "triggerTime_ms": 1,
            "audioEventId": [
              2
            ],
            "volume": 3.0,
            "probability": [
              4.0
            ],
            "hasAlts": true
          }
        ],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.audio_event_ids[0], 2)
        self.assertEqual(keyframe.volume, 3.0)
        self.assertEqual(keyframe.probabilities[0], 4.0)
        self.assertTrue(keyframe.has_alts)

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.audio_event_ids[0], 2)
        self.assertEqual(keyframe.volume, 3.0)
        self.assertEqual(keyframe.probabilities[0], 4.0)
        self.assertTrue(keyframe.has_alts)

        data2 = clips.to_dict()
        self.assertEqual(data, data2)


class TestEvent(unittest.TestCase):

    def test_single(self):
        data = json.loads("""
{
  "clips": [
    {
      "Name": "test",
      "keyframes": {
        "LiftHeightKeyFrame": [],
        "ProceduralFaceKeyFrame": [],
        "HeadAngleKeyFrame": [],
        "RobotAudioKeyFrame": [],
        "BackpackLightsKeyFrame": [],
        "FaceAnimationKeyFrame": [],
        "EventKeyFrame": [
          {
            "triggerTime_ms": 1,
            "event_id": "test"
          }
        ],
        "BodyMotionKeyFrame": [],
        "RecordHeadingKeyFrame": [],
        "TurnToRecordedHeadingKeyFrame": []
      }
    }
  ]
}      
        """)
        clips = AnimClips.from_dict(data)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.event_id, "test")

        f = io.BytesIO()
        clips.to_fb_stream(f)
        f.seek(0, io.SEEK_SET)

        clips = clips.from_fb_stream(f)
        self.assertEqual(len(clips.clips[0].keyframes), 1)
        keyframe = clips.clips[0].keyframes[0]
        self.assertEqual(keyframe.trigger_time_ms, 1)
        self.assertEqual(keyframe.event_id, "test")

        data2 = clips.to_dict()
        self.assertEqual(data, data2)




import unittest

from pycozmo.image_encoder import ImageDecoder, image_to_str
from pycozmo.util import hex_load
from pycozmo.tests.image_encoder_fixtures import FIXTURES


class TestImageDecoder(unittest.TestCase):

    @staticmethod
    def _decode(seq: str) -> ImageDecoder:
        if isinstance(seq, str):
            buffer = hex_load(seq)
        else:
            buffer = bytes(seq)
        decoder = ImageDecoder(buffer)
        decoder.decode()
        return decoder

    def assertSameImage(self, expected: str, decoder: ImageDecoder) -> None:
        actual = image_to_str(decoder.image)
        self.assertEqual(expected.strip(), actual.strip())

    def test_blank(self):
        fixture = FIXTURES["blank"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)
        self.assertEqual(128, decoder.x)
        self.assertEqual(0, decoder.y)

    def test_fill_screen(self):
        fixture = FIXTURES["fill_screen"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)
        self.assertEqual(128, decoder.x)
        self.assertEqual(0, decoder.y)

    def test_fill_screen2(self):
        fixture = FIXTURES["fill_screen2"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)
        self.assertEqual(128, decoder.x)
        self.assertEqual(0, decoder.y)

    def test_top_left(self):
        fixture = FIXTURES["top_left"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_top_left_5(self):
        fixture = FIXTURES["top_left_5"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_top_left_1_8(self):
        fixture = FIXTURES["top_left_1_8"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_top_left_line(self):
        fixture = FIXTURES["top_left_line"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_top_line(self):
        fixture = FIXTURES["top_line"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_bottom_line(self):
        fixture = FIXTURES["bottom_line"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_left_line(self):
        fixture = FIXTURES["left_line"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_right_line(self):
        fixture = FIXTURES["right_line"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_columns(self):
        fixture = FIXTURES["columns"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_rect(self):
        fixture = FIXTURES["rect"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_rect2(self):
        fixture = FIXTURES["rect2"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_rect3(self):
        fixture = FIXTURES["rect3"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_rect4(self):
        fixture = FIXTURES["rect4"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_diagonal(self):
        fixture = FIXTURES["diagonal"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_diagonal2(self):
        fixture = FIXTURES["diagonal2"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_blocks(self):
        fixture = FIXTURES["blocks"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_pycozmo(self):
        fixture = FIXTURES["pycozmo"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard_tl(self):
        fixture = FIXTURES["chessboard_tl"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard_bl(self):
        fixture = FIXTURES["chessboard_bl"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard_tr(self):
        fixture = FIXTURES["chessboard_tr"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard_br(self):
        fixture = FIXTURES["chessboard_br"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard2_tl(self):
        fixture = FIXTURES["chessboard2_tl"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard2_bl(self):
        fixture = FIXTURES["chessboard2_bl"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard2_tr(self):
        fixture = FIXTURES["chessboard2_tr"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)

    def test_chessboard2_br(self):
        fixture = FIXTURES["chessboard2_br"]
        seq = fixture["seq"]
        expected = fixture["image"]
        decoder = self._decode(seq)
        self.assertSameImage(expected, decoder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class LiftHeight(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsLiftHeight(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = LiftHeight()
        x.Init(buf, n + offset)
        return x

    # LiftHeight
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # LiftHeight
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # LiftHeight
    def DurationTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # LiftHeight
    def HeightMm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # LiftHeight
    def HeightVariabilityMm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

def LiftHeightStart(builder): builder.StartObject(4)
def LiftHeightAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def LiftHeightAddDurationTimeMs(builder, durationTimeMs): builder.PrependUint32Slot(1, durationTimeMs, 0)
def LiftHeightAddHeightMm(builder, heightMm): builder.PrependUint8Slot(2, heightMm, 0)
def LiftHeightAddHeightVariabilityMm(builder, heightVariabilityMm): builder.PrependUint8Slot(3, heightVariabilityMm, 0)
def LiftHeightEnd(builder): return builder.EndObject()




from typing import Optional
from abc import ABC, abstractmethod

from .protocol_declaration import PacketType, FIRST_ROBOT_PACKET_ID
from .protocol_utils import BinaryReader, BinaryWriter
from .util import hex_dump


class Struct(ABC):

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def to_writer(self, writer: BinaryWriter) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_bytes(cls, buffer: bytes):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_reader(cls, reader: BinaryReader):
        raise NotImplementedError


class Packet(Struct, ABC):

    __slots__ = (
        "_type",
        "_id",
        "seq",
        "ack",
    )

    def __init__(self, packet_type: PacketType, packet_id: Optional[int] = None):
        self.type = packet_type
        self.id = packet_id
        self.seq = 0
        self.ack = 0

    @property
    def type(self) -> PacketType:
        return self._type

    @type.setter
    def type(self, value: PacketType):
        self._type = PacketType(value)

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, value: Optional[int]):
        self._id = value

    def is_oob(self) -> bool:
        res = self.type.value >= PacketType.EVENT.value
        return res

    def is_from_robot(self) -> bool:
        if self.id is not None:
            res = self.id >= FIRST_ROBOT_PACKET_ID
        else:
            res = self.type == PacketType.CONNECT or (self.type == PacketType.PING and self.seq > 0)
        return res

    def is_from_engine(self) -> bool:
        return not self.is_from_robot()


class UnknownPacket(Packet):

    __slots__ = (
        "_data",
    )

    def __init__(self, packet_type: PacketType, data: bytes, packet_id: Optional[int] = None):
        super().__init__(packet_type, packet_id)
        self.data = data

    @property
    def data(self) -> bytes:
        return self._data

    @data.setter
    def data(self, value: bytes):
        self._data = bytes(value)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return "{type_name}({type:02x}, {data})".format(
            type=self.type.value, type_name=type(self).__name__, data=hex_dump(data=self._data))

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()

    def to_writer(self, writer):
        writer.write_farray(self._data, "B", len(self._data))

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError


class UnknownCommand(UnknownPacket):

    def __init__(self, packet_id: int, data: bytes = b""):
        super().__init__(PacketType.COMMAND, data, packet_id=packet_id)

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError

    def __repr__(self):
        return "{type}({id:02x}, {data})".format(
            id=self.id, type=type(self).__name__, data=hex_dump(data=self._data))


class UnknownEvent(UnknownPacket):

    def __init__(self, packet_id: int, data: bytes = b""):
        super().__init__(PacketType.EVENT, data, packet_id=packet_id)

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError

    def __repr__(self):
        return "{type}({id:02x}, {data})".format(
            id=self.id, type=type(self).__name__, data=hex_dump(data=self._data))




from typing import Optional, Tuple
import math


__all__ = [
    'Angle',
    'Distance',
    'Speed',
    'Vector3',
    'angle_z_to_quaternion',
    'Matrix44',
    'Quaternion',
    'Pose',

    'hex_dump',
    'hex_load',
    'frange',
]


class Angle:
    """
    Angle representation.

    Args:
        radians (float): The number of radians the angle should represent
            (cannot be combined with ``degrees``)
        degrees (float): The number of degress the angle should represent
            (cannot be combined with ``radians``)
    """

    __slots__ = '_radians'

    def __init__(self, radians: Optional[float] = None, degrees: Optional[float] = None):
        if radians is None and degrees is None:
            raise ValueError("Expected either the degrees or radians keyword argument")
        if radians and degrees:
            raise ValueError("Expected either the degrees or radians keyword argument, not both")

        if degrees is not None:
            radians = degrees * math.pi / 180
        self._radians = float(radians)

    def __repr__(self):
        return "<%s %.2f radians (%.2f degrees)>" % (self.__class__.__name__, self.radians, self.degrees)

    def __add__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported type for + expected Angle")
        return Angle(radians=self.radians + other.radians)

    def __sub__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported type for - expected Angle")
        return Angle(radians=self.radians - other.radians)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported type for * expected number")
        return Angle(radians=self.radians * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported type for / expected number")
        return Angle(radians=self.radians / other)

    def _cmp_int(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported type for comparison expected Angle")
        return self.radians - other.radians

    def __eq__(self, other):
        return self._cmp_int(other) == 0

    def __ne__(self, other):
        return self._cmp_int(other) != 0

    def __gt__(self, other):
        return self._cmp_int(other) > 0

    def __lt__(self, other):
        return self._cmp_int(other) < 0

    def __ge__(self, other):
        return self._cmp_int(other) >= 0

    def __le__(self, other):
        return self._cmp_int(other) <= 0

    @property
    def radians(self) -> float:
        """ Returns the angle in radians. """
        return self._radians

    @property
    def degrees(self) -> float:
        """ Returns the angle in degrees. """
        return self._radians / math.pi * 180

    @property
    def abs_value(self):
        """:class:`cozmo.util.Angle`: The absolute value of the angle.

        If the Angle is positive then it returns a copy of this Angle, otherwise it returns -Angle.
        """
        return Angle(radians=abs(self._radians))


class Distance:
    """
    Represents a distance.

    The class allows distances to be returned in either millimeters or inches.

    Args:
        mm (float): The number of millimeters the distance should
            represent (cannot be combined with ``distance_inches``).
        inches (float): The number of inches the distance should
            represent (cannot be combined with ``distance_mm``).
    """

    __slots__ = '_mm'

    def __init__(self, mm: Optional[float] = None, inches: Optional[float] = None):
        if mm is None and inches is None:
            raise ValueError("Expected either the mm or inches keyword argument")
        if mm and inches:
            raise ValueError("Expected either the mm or inches keyword argument, not both")

        if inches is not None:
            mm = inches * 25.4
        self._mm = mm

    def __repr__(self):
        return "<%s %.2f mm (%.2f inches)>" % (self.__class__.__name__, self.mm, self.inches)

    def __add__(self, other):
        if not isinstance(other, Distance):
            raise TypeError("Unsupported operand for + expected Distance")
        return Distance(mm=self.mm + other.mm)

    def __sub__(self, other):
        if not isinstance(other, Distance):
            raise TypeError("Unsupported operand for - expected Distance")
        return Distance(mm=self.mm - other.mm)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return Distance(mm=self.mm * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return Distance(mm=self.mm / other)

    @property
    def mm(self) -> float:
        """ The distance in millimeters. """
        return self._mm

    @property
    def inches(self) -> float:
        """ The distance in inches. """
        return self._mm / 25.4


class Speed:
    """
    Speed representation.

    Args:
        mmps (float): The number of millimeters per second the speed should represent.
    """

    __slots__ = '_mmps'

    def __init__(self, mmps: float):
        self._mmps = mmps

    def __repr__(self):
        return "<%s %.2f mmps>" % (self.__class__.__name__, self.mmps)

    def __add__(self, other):
        if not isinstance(other, Speed):
            raise TypeError("Unsupported operand for + expected Speed")
        return Speed(mmps=self.mmps + other.mmps)

    def __sub__(self, other):
        if not isinstance(other, Speed):
            raise TypeError("Unsupported operand for - expected Speed")
        return Speed(mmps=self.mmps - other.mmps)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return Speed(mmps=self.mmps * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return Speed(mmps=self.mmps / other)

    @property
    def mmps(self) -> float:
        """ Returns the speed in millimeters per second (mmps). """
        return self._mmps


class Vector3:
    """
    Represents a 3D Vector (type/units aren't specified).

    Args:
        x (float): X component
        y (float): Y component
        z (float): Z component
    """

    __slots__ = ('_x', '_y', '_z')

    def __init__(self, x: float, y: float, z: float):
        self._x = x
        self._y = y
        self._z = z

    def set_to(self, rhs):
        """
        Copy the x, y and z components of the given vector.

        Args:
            rhs (:class:`Vector3`): The right-hand-side of this assignment - the
                source vector to copy into this vector.
        """
        self._x = rhs.x
        self._y = rhs.y
        self._z = rhs.z

    @property
    def x(self) -> float:
        """ The x component. """
        return self._x

    @property
    def y(self) -> float:
        """ The y component. """
        return self._y

    @property
    def z(self) -> float:
        """ The z component. """
        return self._z

    @property
    def x_y_z(self) -> Tuple[float, float, float]:
        """ The X, Y, Z elements of the Vector3 (x,y,z). """
        return self._x, self._y, self._z

    def __repr__(self):
        return "<%s x: %.2f y: %.2f z: %.2f>" % (self.__class__.__name__, self.x, self.y, self.z)

    def __add__(self, other):
        if not isinstance(other, Vector3):
            raise TypeError("Unsupported operand for + expected Vector3")
        return Vector3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        if not isinstance(other, Vector3):
            raise TypeError("Unsupported operand for - expected Vector3")
        return Vector3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return Vector3(self.x * other, self.y * other, self.z * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return Vector3(self.x / other, self.y / other, self.z / other)


def angle_z_to_quaternion(angle_z: Angle) -> Tuple[float, float, float, float]:
    """
    Converts an angle in the z axis (Euler angle z component) to a quaternion.
    """
    # Define the quaternion to be converted from a Euler angle (x,y,z) of 0,0,angle_z
    # These equations have their original equations above, and simplified implemented
    # q0 = cos(x/2)*cos(y/2)*cos(z/2) + sin(x/2)*sin(y/2)*sin(z/2)
    q0 = math.cos(angle_z.radians / 2)
    # q1 = sin(x/2)*cos(y/2)*cos(z/2) - cos(x/2)*sin(y/2)*sin(z/2)
    q1 = 0
    # q2 = cos(x/2)*sin(y/2)*cos(z/2) + sin(x/2)*cos(y/2)*sin(z/2)
    q2 = 0
    # q3 = cos(x/2)*cos(y/2)*sin(z/2) - sin(x/2)*sin(y/2)*cos(z/2)
    q3 = math.sin(angle_z.radians / 2)
    return q0, q1, q2, q3


class Matrix44:
    """
    A 4x4 Matrix for representing the rotation and/or position of an object in the world.

    Can be generated from a Quaternion for a pure rotation matrix, or
    combined with a position for a full translation matrix, as done by Pose.to_matrix().
    """
    __slots__ = ('m00', 'm10', 'm20', 'm30',
                 'm01', 'm11', 'm21', 'm31',
                 'm02', 'm12', 'm22', 'm32',
                 'm03', 'm13', 'm23', 'm33')

    def __init__(self,
                 m00, m10, m20, m30,
                 m01, m11, m21, m31,
                 m02, m12, m22, m32,
                 m03, m13, m23, m33):
        self.m00 = m00
        self.m10 = m10
        self.m20 = m20
        self.m30 = m30

        self.m01 = m01
        self.m11 = m11
        self.m21 = m21
        self.m31 = m31

        self.m02 = m02
        self.m12 = m12
        self.m22 = m22
        self.m32 = m32

        self.m03 = m03
        self.m13 = m13
        self.m23 = m23
        self.m33 = m33

    def __repr__(self):
        return ("<%s: "
                "%.1f %.1f %.1f %.1f %.1f %.1f %.1f %.1f "
                "%.1f %.1f %.1f %.1f %.1f %.1f %.1f %.1f>" % (self.__class__.__name__, *self.in_row_order))

    @property
    def tabulated_string(self):
        """str: A multi-line string formatted with tabs to show the matrix contents."""
        return ("%.1f\t%.1f\t%.1f\t%.1f\n"
                "%.1f\t%.1f\t%.1f\t%.1f\n"
                "%.1f\t%.1f\t%.1f\t%.1f\n"
                "%.1f\t%.1f\t%.1f\t%.1f" % self.in_row_order)

    @property
    def in_row_order(self) -> Tuple[float, float, float, float,
                                    float, float, float, float,
                                    float, float, float, float,
                                    float, float, float, float]:
        """
        Returns the contents of the matrix in row order.
        """
        return self.m00, self.m01, self.m02, self.m03, \
            self.m10, self.m11, self.m12, self.m13, \
            self.m20, self.m21, self.m22, self.m23, \
            self.m30, self.m31, self.m32, self.m33

    @property
    def in_column_order(self) -> Tuple[float, float, float, float,
                                       float, float, float, float,
                                       float, float, float, float,
                                       float, float, float, float]:
        """
        Returns the contents of the matrix in column order.
        """
        return self.m00, self.m10, self.m20, self.m30, \
            self.m01, self.m11, self.m21, self.m31, \
            self.m02, self.m12, self.m22, self.m32, \
            self.m03, self.m13, self.m23, self.m33

    @property
    def forward_xyz(self) -> Tuple[float, float, float]:
        """
        Returns the x,y,z components representing the matrix's forward vector.
        """
        return self.m00, self.m01, self.m02

    @property
    def left_xyz(self) -> Tuple[float, float, float]:
        """
        Returns the x,y,z components representing the matrix's left vector.
        """
        return self.m10, self.m11, self.m12

    @property
    def up_xyz(self) -> Tuple[float, float, float]:
        """
        Returns the x,y,z components representing the matrix's up vector.
        """
        return self.m20, self.m21, self.m22

    @property
    def pos_xyz(self) -> Tuple[float, float, float]:
        """
        Returns the x,y,z components representing the matrix's position vector.
        """
        return self.m30, self.m31, self.m32

    def set_forward(self, x: float, y: float, z: float):
        """
        Set the x,y,z components representing the matrix's forward vector.
        """
        self.m00 = x
        self.m01 = y
        self.m02 = z

    def set_left(self, x: float, y: float, z: float):
        """
        Set the x,y,z components representing the matrix's left vector.
        """
        self.m10 = x
        self.m11 = y
        self.m12 = z

    def set_up(self, x: float, y: float, z: float):
        """
        Set the x,y,z components representing the matrix's up vector.
        """
        self.m20 = x
        self.m21 = y
        self.m22 = z

    def set_pos(self, x: float, y: float, z: float) -> None:
        """
        Set the x,y,z components representing the matrix's position vector.
        """
        self.m30 = x
        self.m31 = y
        self.m32 = z


class Quaternion:
    """
    Represents rotation.
    """

    __slots__ = ('_q0', '_q1', '_q2', '_q3')

    def __init__(self,
                 q0: Optional[float] = None,
                 q1: Optional[float] = None,
                 q2: Optional[float] = None,
                 q3: Optional[float] = None,
                 angle_z: Optional[Angle] = None):

        is_quaternion = (q0 is not None) and (q1 is not None) and (q2 is not None) and (q3 is not None)

        if not is_quaternion and angle_z is None:
            raise ValueError("Expected either the q0 q1 q2 and q3 or angle_z keyword arguments")
        if is_quaternion and angle_z:
            raise ValueError("Expected either the q0 q1 q2 and q3 or angle_z keyword argument, not both")
        if angle_z is not None:
            if not isinstance(angle_z, Angle):
                raise TypeError("Unsupported type for angle_z expected Angle")
            q0, q1, q2, q3 = angle_z_to_quaternion(angle_z)

        self._q0 = q0
        self._q1 = q1
        self._q2 = q2
        self._q3 = q3

    def __repr__(self):
        return ("<%s q0: %.2f q1: %.2f q2: %.2f q3: %.2f (angle_z: %s)>" %
                (self.__class__.__name__, self.q0, self.q1, self.q2, self.q3, self.angle_z))

    def to_matrix(self, pos_x: float = 0.0, pos_y: float = 0.0, pos_z: float = 0.0):
        """
        Convert the Quaternion to a 4x4 matrix representing this rotation.

        A position can also be provided to generate a full translation matrix.
        """

        # See https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
        q0q0 = self.q0 * self.q0
        q1q1 = self.q1 * self.q1
        q2q2 = self.q2 * self.q2
        q3q3 = self.q3 * self.q3

        q0x2 = self.q0 * 2.0  # saves 2 multiplies
        q0q1x2 = q0x2 * self.q1
        q0q2x2 = q0x2 * self.q2
        q0q3x2 = q0x2 * self.q3
        q1x2 = self.q1 * 2.0  # saves 1 multiply
        q1q2x2 = q1x2 * self.q2
        q1q3x2 = q1x2 * self.q3
        q2q3x2 = 2.0 * self.q2 * self.q3

        m00 = (q0q0 + q1q1 - q2q2 - q3q3)
        m01 = (q1q2x2 + q0q3x2)
        m02 = (q1q3x2 - q0q2x2)

        m10 = (q1q2x2 - q0q3x2)
        m11 = (q0q0 - q1q1 + q2q2 - q3q3)
        m12 = (q0q1x2 + q2q3x2)

        m20 = (q0q2x2 + q1q3x2)
        m21 = (q2q3x2 - q0q1x2)
        m22 = (q0q0 - q1q1 - q2q2 + q3q3)

        return Matrix44(m00, m10, m20, pos_x,
                        m01, m11, m21, pos_y,
                        m02, m12, m22, pos_z,
                        0.0, 0.0, 0.0, 1.0)

    # These are only for angle_z because quaternion addition/subtraction is not relevant here
    def __add__(self, other):
        if not isinstance(other, Quaternion):
            raise TypeError("Unsupported operand for + expected Quaternion")
        angle_z = self.angle_z + other.angle_z
        return Quaternion(angle_z=angle_z)

    def __sub__(self, other):
        if not isinstance(other, Quaternion):
            raise TypeError("Unsupported operand for - expected Quaternion")
        angle_z = self.angle_z - other.angle_z
        return Quaternion(angle_z=angle_z)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        angle_z = self.angle_z * other
        return Quaternion(angle_z=angle_z)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        angle_z = self.angle_z / other
        return Quaternion(angle_z=angle_z)

    @property
    def q0(self) -> float:
        return self._q0

    @property
    def q1(self) -> float:
        return self._q1

    @property
    def q2(self) -> float:
        return self._q2

    @property
    def q3(self) -> float:
        return self._q3

    @property
    def q0_q1_q2_q3(self) -> Tuple[float, float, float, float]:
        return self._q0, self._q1, self._q2, self._q3

    @property
    def angle_z(self) -> Angle:
        radians = math.atan2(2 * (self.q1 * self.q2 + self.q0 * self.q3),
                             1 - 2 * (self.q2 ** 2 + self.q3 ** 2))
        return Angle(radians=radians)

    @property
    def euler_angles(self) -> Tuple[float, float, float]:
        """
        Returns the pitch, yaw, roll Euler components of the object's
        rotation defined as rotations in the x, y, and z axis respectively.

        :return:
        """

        # convert to matrix
        matrix = self.to_matrix()

        # normalize the magnitudes of cos(roll)*sin(pitch) (i.e. m12) and
        #   cos(roll)*cos(pitch) (ie. m22), to isolate cos(roll) to be compared
        #   against -sin(roll) (m02).  Unfortunately, this omits results with an
        #   absolute angle larger than 90 degrees on roll.
        absolute_cos_roll = math.sqrt(matrix.m12 * matrix.m12 + matrix.m22 * matrix.m22)
        near_gimbal_lock = absolute_cos_roll < 1e-6
        if not near_gimbal_lock:
            # general case euler decomposition
            pitch = math.atan2(matrix.m22, matrix.m12)
            yaw = math.atan2(matrix.m00, matrix.m01)
            roll = math.atan2(absolute_cos_roll, -matrix.m02)
        else:
            # special case euler angle decomposition near gimbal lock
            pitch = math.atan2(matrix.m11, -matrix.m21)
            yaw = 0
            roll = math.atan2(absolute_cos_roll, -matrix.m02)

        # adjust roll to be consistent with how the device is oriented
        roll = math.pi * 0.5 - roll
        if roll > math.pi:
            roll -= math.pi * 2

        return pitch, yaw, roll


class Pose:
    """
    A combination of position (vector) and rotation (quaternion).
    """

    __slots__ = ('_position', '_rotation', '_origin_id', '_is_accurate')

    def __init__(self,
                 x: float, y: float, z: float,
                 q0: Optional[float] = None, q1: Optional[float] = None,
                 q2: Optional[float] = None, q3: Optional[float] = None,
                 angle_z: Optional[Angle] = None, origin_id: int = -1, is_accurate: bool = True):
        self._position = Vector3(x, y, z)
        self._rotation = Quaternion(q0, q1, q2, q3, angle_z)
        self._origin_id = origin_id
        self._is_accurate = is_accurate

    @classmethod
    def _create_from_clad(cls, pose):
        return cls(pose.x, pose.y, pose.z,
                   q0=pose.q0, q1=pose.q1, q2=pose.q2, q3=pose.q3,
                   origin_id=pose.originID)

    @classmethod
    def _create_default(cls):
        return cls(0.0, 0.0, 0.0,
                   q0=1.0, q1=0.0, q2=0.0, q3=0.0,
                   origin_id=-1)

    def __repr__(self):
        return "<%s %s %s origin_id=%d>" % (self.__class__.__name__, self.position, self.rotation, self.origin_id)

    def __add__(self, other):
        if not isinstance(other, Pose):
            raise TypeError("Unsupported operand for + expected Pose")
        pos = self.position + other.position
        rot = self.rotation + other.rotation
        return Pose(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __sub__(self, other):
        if not isinstance(other, Pose):
            raise TypeError("Unsupported operand for - expected Pose")
        pos = self.position - other.position
        rot = self.rotation - other.rotation
        return Pose(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        pos = self.position * other
        rot = self.rotation * other
        return Pose(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        pos = self.position / other
        rot = self.rotation / other
        return Pose(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def define_pose_relative_this(self, new_pose):
        """
        Creates a new pose such that new_pose's origin is now at the location of this pose.
        """

        if not isinstance(new_pose, Pose):
            raise TypeError("Unsupported type for new_origin, must be of type Pose")
        x, y, z = self.position.x_y_z
        angle_z = self.rotation.angle_z
        new_x, new_y, new_z = new_pose.position.x_y_z
        new_angle_z = new_pose.rotation.angle_z

        cos_angle = math.cos(angle_z.radians)
        sin_angle = math.sin(angle_z.radians)
        res_x = x + (cos_angle * new_x) - (sin_angle * new_y)
        res_y = y + (sin_angle * new_x) + (cos_angle * new_y)
        res_z = z + new_z
        res_angle = angle_z + new_angle_z
        return Pose(res_x, res_y, res_z, angle_z=res_angle, origin_id=self._origin_id)

    def invalidate(self) -> None:
        """
        Mark this pose as being invalid (unusable).
        """
        self._origin_id = -1

    def is_comparable(self, other_pose) -> bool:
        """
        Are these two poses comparable.

        Poses are comparable if they're valid and having matching origin IDs.
        """
        return (self.is_valid and other_pose.is_valid and
                (self.origin_id == other_pose.origin_id))

    @property
    def is_valid(self) -> bool:
        """
        Checks whether a pose is valid (usable).
        """
        return self.origin_id >= 0

    @property
    def position(self) -> Vector3:
        """
        Returns the position component of this pose.
        """
        return self._position

    @property
    def rotation(self) -> Quaternion:
        """
        Returns the rotation component of this pose.
        """
        return self._rotation

    def to_matrix(self):
        """
        Convert the Pose to a Matrix44.
        """
        return self.rotation.to_matrix(*self.position.x_y_z)

    @property
    def origin_id(self) -> int:
        """
        Returns an ID maintained by the robot (engine) which represents which coordinate frame this pose is in.
        """
        return self._origin_id

    @origin_id.setter
    def origin_id(self, value: int) -> None:
        """
        Change the ID of a pose.
        """
        if not isinstance(value, int):
            raise TypeError("The type of origin_id must be int")
        self._origin_id = value

    @property
    def is_accurate(self) -> bool:
        """
        Returns True if this pose is valid and accurate.

        Poses are marked as inaccurate if we detect movement via accelerometer,
        or if they were observed from far enough away that we're less certain
        of the exact pose.
        """
        return self.is_valid and self._is_accurate


def hex_dump(data: bytes) -> str:
    res = ":".join("{:02x}".format(b) for b in data)
    return res


def hex_load(data: str) -> bytes:
    res = bytearray.fromhex(data.replace(":", ""))
    return res


def frange(start, stop, step):
    x = start
    while x < stop:
        yield x
        x += step
